:- module(lexer, [scanTokens/2, scanFile/2, display_tokens/1], _).

:- use_module(library(lists),         [length/2] ).
:- use_module(library(dynamic),       [assert/1, retract/1, retractall/1] ).
:- use_module(library(file_utils),    [file_to_string/2]).
:- use_module(library(iso_byte_char), [atom_chars/2,char_code/2]).

:- use_package(hiord).

% egallego@babel: 02 Nov 2004, 17:09
% 
% Lexer module. It exports two predicates, scanTokens and scanFile 
% obtaining a list of tokens.
% 
% If you want to use old interface it is in lexer_compat module.

% Wrapper to scan from a file.
scanFile(File, Tokens) :-
	file_to_string(File, S),
	scanTokens(S, Tokens).

% scanTokens(String, ListaTokens) se encarga de tokenizar un string.
% de momento lee de un String, pero puede leer de 

scanTokens(String, ListaTokens) :-
	init_pos,
	init_layout,
	scanTokensList(String, ListaTokens).

% This is the main lexer function.
scanTokensList([], []) :- !.
scanTokensList(Input, [Token|Tokens]) :-
	scanToken(Input, NewInput, Token),
	!,
%	process_layout(Token, Tokens),
	scanTokensList(NewInput, Tokens).

% scanToken (Input, NewInput, Token)
scanToken([],[],[]).
scanToken(Input, NewInput, Token) :-
	filter_blanks(Input, TInput),
	(
	    (
		is_bol
	    ;
		is_new_layout_word
	    ),
	    \+ TInput = [] ->
	    get_layout_token(TInput, NewInput, Token)
	;
	    get_real_token(TInput, NewInput, Token)
	).

% PRE: This predicate should be called with no blanks in Input.
get_real_token([], [], T) :- 
	set_actual_pos(eof, T),
	!.
get_real_token(Input, NewInput, Token) :-
	select_viable_token(Input, NewInput, CandToken),
	(
	    token(CandToken, NToken) ->
	    actual_pos(Line, Col),
	    Token = (NToken, pos(Line, Col)),
	    update_pos(CandToken),
	    tag_special_layout_token(NToken)
	;
	    actual_pos(Line, _),
	    throw(lexer_error(Line, CandToken, Input, NewInput))
	).

% select_viable_token takes as much of its input as it
% can and returns it.
% In the future this should use regular expressions.
select_viable_token([],[],[]).

% Identifiers.
select_viable_token([C|CH], NewInput, CToken) :-
	is_alfa(C), 
	!,
	get_while(is_alfanum ,[C|CH], NewInput, CToken).

% Numbers.
select_viable_token([C|CH], NewInput, CToken) :-
	is_digit(C),
	!,
	get_while(is_alfanum, [C|CH], NewInput, CToken).

select_viable_token([CToken|CH], CH, [CToken]) :-
	is_punt(CToken), 
	!.

select_viable_token(Input, NewInput, CToken) :-
	get_while(is_not_alfanum, Input, NewInput, CToken), 
	!.

% A high order get_while.

:- meta_predicate get_while(pred(1), ?, ?, ?).

get_while(Goal, [C|CH], ListRes, [C|RS]) :-
	Goal(C),
	!,
	get_while(Goal, CH, ListRes, RS).

get_while(_, List, List, []).


% Layout processing.

:- dynamic layout_props/1.

is_bol :-
	layout_props(bol).

is_new_layout_word :-
	layout_props(special_word).

init_layout :-
	retractall(layout_props(_)),
	assert(layout_props(layout([1]))). %FIXME

layout_pos(C) :-
	layout_props(layout([C|_])).
	
pushLayout(N) :-
	retract(layout_props(layout(L))),
	assert(layout_props(layout([N|L]))).

popLayout :-
	retract(layout_props(layout([_|L]))),
	assert(layout_props(layout(L))).

tag_special_layout_token(reserved(where)) :-
	assert(layout_props(special_word)).

tag_special_layout_token(reserved(let)) :-
	assert(layout_props(special_word)).

tag_special_layout_token(reserved(of)) :-
	assert(layout_props(special_word)).

tag_special_layout_token(reserved(do)) :-
	assert(layout_props(special_word)).

tag_special_layout_token(_).


set_actual_pos(Symbol, Token) :-
	actual_pos(L, C),
	Token = (Symbol, pos(L, C)).

% This predicate returns the new a layout token if it's needed or
% a real token otherwise.
get_layout_token(Input, NewInput, Token) :-
	(
	    is_bol ->
	    retract(layout_props(bol)),
	    actual_pos(_, CT),
	    layout_pos(CL),
	    (
		CT = CL ->
		set_actual_pos(';', Token),
		Input = NewInput	    
	    ;
		(
		    CT < CL ->
		    popLayout,
		    set_actual_pos('}', Token),
		    Input = NewInput
		;
		    get_real_token(Input, NewInput, Token)
		)
	    )
	;
	    is_new_layout_word ->
	    retract(layout_props(special_word)),
	    actual_pos(_, CT),
	    pushLayout(CT),
	    set_actual_pos('{', Token),
	    Input = NewInput
	).

% Tipos de tokens que tenemos: 
%
is_token(X) :-
	token(_, X), !.

 
% op(X) -> Indica un operador.
token("..",  op(..)).
token(":=",  op(:=)).
token("=>",  op(=>)).
token(">>",  op(>>)).
token(">>=", op(>>=)).
token("=",   op(=)).
token("==",  op(==)).
token("===", op(===)).
token("@",   op(@)).
token("\\",  op(backslash)).
% I'm sure | is a punt sign.
token("|",   op(pipe)).
token("<-",  op(<-)).
token("->",  op(->)).
token("~",   op(tilde)).
token("-",   op(-)).
token("+",   op(+)).
token("*",   op(*)).
token("/",   op(/)).
token(">",   op(>)).
token("<",   op(>)).
token(">=",  op(>=)).
token("=<",  op(=<)).
token("&&",  op(&&)).
token("/\\", op(/\)).
token("::",  op(::)).
token(":",   op(:)).

% Punctuation tokens.

token("(", '(').
token(")", ')').
token("[", '[').
token("]", ']').
token("{", '{').
token("}", '}').
token(",", ',').
token(";", ';').
token(".", '.').
% I think _ is not a punct token.
%token("_", '_').

% Reserved words.
token(Name, reserved(Token)) :-
	atom_codes(Token, Name),
	is_reserved(Token).

% numbers.
token(X, num(N)) :-
	number_codes(N, X).

% identifier tokens.
token([C|CH], id(N)) :-
	is_lower(C),
	atom_codes(N, [C|CH]).

token([C|CH], conid(N)) :-
	is_upper(C),
	atom_codes(N, [C|CH]).

% String token.
% Catch errors.
%token(X, string(X)).



is_punt(C) :-
	char_code(T, C),
	member(T, ['{','}','[',']','(',')',',','.',';']).

is_reserved(I) :-
	member(I,[and,eval,class,data,default,deriving,do,else,hiding,
	          if,import,in,infix,infixl,infixr,instance,let,
		  module,newtype,of,or,qualified,rule,then,type,
		  where,rigid,flex,curry_eval,choice,
		  ord,chr,'String','Char','Bool','Int','Float']).


% Various utils.
% FIXME.
% In filter blanks we should follow the layout.
filter_blanks([],[]).
filter_blanks([0'-, 0'-|CS], NS) :-
	!,
	inc_pos(2),
	skip_comment(CS, NCS),
	filter_blanks(NCS, NS).

filter_blanks([C|CS], NS) :-
	is_blank(C), 
	skip_blank(C),
	!,
	filter_blanks(CS, NS).

filter_blanks(X, X).

% 
skip_blank(0' ) :-
	inc_pos(1).

skip_blank(0'\n) :-
	new_line.

% Following Haskell Layout.
skip_blank(0'\t) :-
	inc_pos(8).

skip_comment([C|CH], NS) :-
	C \== 0'\n, 
	!,
	inc_pos(1),
	skip_comment(CH, NS).

skip_comment(X, X).

% Position handlers.
% Now we need to follow the position in the parser to track layout.

% actual_pos(line, column)
:- dynamic actual_pos/2.

init_pos :-
	(
	    actual_pos(_,_) ->
	    retract(actual_pos(_,_))
	;
	    true
	),
	assert(actual_pos(1, 1)).

inc_pos(N) :-
	retract(actual_pos(Line, Column)),
	NColumn is Column + N,
	assert(actual_pos(Line, NColumn)).

new_line :-
	retract(actual_pos(Line, _)),
	NLine is Line + 1,
	assert(actual_pos(NLine, 1)),
	(
	    \+layout_props(bol) ->
	    assert(layout_props(bol))
	;
	    true
	).

update_pos(String) :-
	length(String, N),
	inc_pos(N).

% FIXME: Of moment we use \n to mark end of rule, but this should
% be done using layout.

is_blank(X) :-
	member(X," \n\t").

is_alfanum(CH) :-
	is_alfa(CH).
is_alfanum(CH) :-
	is_digit(CH).

is_not_alfanum(CH) :-
	\+ is_alfanum(CH),
	\+ is_blank(CH),
	\+ is_punt(CH).

is_alfa(CH) :-
	is_upper(CH).
is_alfa(CH) :-
	is_lower(CH).

is_upper(N):-
	N>=65,
	N=<90.

is_lower(N):-
	(N>=0'a,N=<0'z ; N=0'_),!.
is_lower(N):-
	N>127,!.

is_digit(N):-
	N<58,
	N>47.


% Pretty print function.
display_tokens([]) :- !.

display_tokens([(T, _)|TS]) :-
	display_token(T),
	display(' '),
	!,
	display_tokens(TS).

display_token(id(X)) :-
	!,
	display(X).

display_token(num(X)) :-
	!,
	display(X).

display_token(conid(X)) :-
	!,
	display(X).

display_token(reserved(X)) :-
	!,
	display(X).

display_token(op(X)) :-
	!,
	display(X).

display_token(X) :-
	!,
	display(X).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

